#ifndef ADDSPECIES_H
#define ADDSPECIES_H

#ifdef __unix__
#define LINUX
#elif defined(WIN32) || defined(_WIN32) || defined(_WIN64)
#define WINDOWS
#endif

#include <iostream>
#include <fstream>
#include <sstream>
#include <iomanip>
#include <string>
#include <vector>
#include <map>
#include <algorithm>
#include <stdlib.h>
#include <time.h>

using namespace std;

#ifndef __max
#define __max(a,b)  (((a) > (b)) ? (a) : (b))
#endif
#ifndef __min
#define __min(a,b)  (((a) < (b)) ? (a) : (b))
#endif

const string version    = "0.8";
const string def_config = "config.ini";
const double def_E      = 1e-10;
const double big_E      = 1e+10;
const double lower_E    = 50.0;     // to lower Elimit for self scores
const double def_pweight = 0.5;
const double def_oweight = 0.5;
const int def_quit      = -1;
const int def_genestep  = 10;
const int buflen        = 65536;
const int ncol          = 2;        // since only protein and gene IDs are needed
const int dummy_score   = 100;      // score value for dummy arcs inserted 
enum Required { field_protein, field_gene };
const int inrecordstep  = 1000;
const int inmatrixstep  = 100000;
const int outmatrixstep = 10000;
const int outgenestep   = 10;
const int maxint        = 0x7ffffff;
const string back9      = "\b\b\b\b\b\b\b\b\b";

struct Gene;
typedef Gene* PGene;
typedef map <string, Gene>::iterator Giter;
typedef pair <string, Gene> Gpair;
typedef map <string, PGene>::iterator Piter;
typedef pair <string, PGene> Ppair;

struct Homolog 
{
    PGene   pgene;
    double  evalue;     // set bigger to disable
    int     score;
    Homolog() { }
    Homolog(PGene pg, double e, int s) {
        pgene  = pg;
        evalue = e;
        score  = s;
    }
    Homolog(const Homolog& h) {
        pgene  = h.pgene;
        evalue = h.evalue;
        score  = h.score;
    }
};

struct Species 
{
    string      name;
    map <string, Gene> genes;
    vector <string> fields;         // Names of fields that present at input
    vector <int> required;          // Maps to a number of required field or -1
    vector <int> myfields;          // [>=ncol] Actual numbers of required and known fields
    Species() {}
    Species(string& s) { name = s; }
};

struct Global 
{
    int         nAnimals;           // Species in total
    int         fromsp;             // source species no.
    int         tosp;               // target species no.
    double      Elimit;             // max E-value to consider
    double      pweight;            // min weight for paralogy
    double      oweight;            // min weight for orthology
    bool        givenElimit;        // if Elimit given in cmd line
    bool        givenPweight;       // if pweight given in cmd line
    bool        givenOweight;       // if oweight given in cmd line
    int         genestep;           // gene no. output step
    bool        outmatrix;          // true: output gene matrix
    bool        unsorted;           // true: don't sort matrix at output
    int         quit;               // quit after: 0=write matrices 1=pass1 2=pass2
    int         add2para;           // add to paralogs: 0=none 1=best hit 2=all
    string      source;             // name of new species
    string      target;             // name of existing species
    string      config;             // name of configuration file
    string      info_path;          // directory of gene tables
    string      info_ext;           // extension of gene tables
    string      homology_path;      // directory of ortholog tables
    string      homology_ext;       // extension of ortholog tables
    string      paralogs_path;      // directory of paralog tables
    string      paralogs_ext;       // extension of paralog tables
    string      matrix_path;        // directory of score matrices
    string      matrix_ext;         // extension of score matrices
    string      matrix_separ;       // separator of score matrices
    string      outmatrix_path;     // directory for matrix output 
    string      outmatrix_ext;      // extension for matrix output
    string      outmatrix_separ;    // separator for matrix output
    string      addition_file;      // separate file for additions
    vector <string>  myfields;      // [>=ncol] Names of required fields in fixed order 
    vector <Species> Animals;       // Array of species, #0=base
    map <string, PGene> stableid;   // Ensembl stable ID -> *Gene mapping
    Global() 
    {
        nAnimals    = 0;
        fromsp      = -1;
        tosp        = -1;
        config      = def_config;
        Elimit      = def_E;
        pweight     = def_pweight;
        oweight     = def_oweight;
        genestep    = def_genestep;
        outmatrix   = false;
        unsorted    = false;
        quit        = def_quit;
        givenElimit  = false;
        givenPweight = false;
        givenOweight = false;
    }
};

#ifdef ADDSPECIES_MAIN
    Global global;
#else
    extern Global global;
#endif      //#ifdef ADDSPECIES_MAIN

struct Gene 
{
    string  name;
    int     sp;                     // animal no. - back reference
    vector <vector <Homolog> > homologs;    // [nAnimals][] incl. paralogs
    //NB: for genes of the target species, add one more vector to obtain [nAnimals+1] ! 
    Gene() 
    {
        sp = -1;
        for (int k = 0; k < global.nAnimals; k++) 
        {
            vector <Homolog> vpg;
            homologs.push_back(vpg);
        }
    }
    ~Gene() 
    {
        homologs.clear();
    }
};

bool    ReadConfig(void);
bool    ReadGeneInfo(int sp);
bool    ReadShortData(int sp, bool paralogs=false);
bool    ReadMatrix(int sp1, int sp2);
bool    WriteMatrix(int sp1, int sp2);
bool    LessID(const PGene p1, const PGene p2);
#endif      //#ifndef ADDSPECIES_H